import numpy as np
cimport numpy as np
cimport cython
cimport numpy.random

from scipy.spatial.distance import cdist

# noinspection PyUnresolvedReferences
from vose cimport Sampler
from libc.stdlib cimport rand, RAND_MAX


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nogil(True)
cdef double sqeuclidean_distance_points(double[::1] x, double[::1] y):
    cdef int dim = y.shape[0]
    cdef double tmp, d
    d = 0.0
    for k in range(dim):
        tmp = x[k] - y[k]
        d += tmp * tmp
    return d


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.nogil(True)
cdef double min_sqeuclidean_distance(double[:, ::1] X, double[::1] y):
    cdef int M = X.shape[0]
    cdef int dim = y.shape[0]
    cdef double d, min_distance
    min_distance = 1e5
    for i in range(M):
        d = sqeuclidean_distance_points(X[i, :], y)
        if d < min_distance:
            min_distance = d
    return min_distance


@cython.boundscheck(False)
@cython.wraparound(False)
@cython.cdivision(True)
def kmc2(X, k, chain_length):
    """Cython implementation of k-MC2 and AFK-MC2 seeding

    Args:
      X: (n,d)-shaped np.ndarray with data points (or scipy CSR matrix)
      k: number of cluster centers
      chain_length: length of the MCMC chain
    Returns:
      (k, d)-shaped numpy.ndarray with cluster centers
    """
    # Local cython variables
    cdef double x_proba, y_proba, distance_to_x, distance_to_y, acceptance_ratio
    cdef double[::1] x_position, y_position, rand_a #, distance_to_first
    cdef np.intp_t x_id, y_id

    # Initialize result
    centers = np.zeros((k, X.shape[1]), np.float64, order="C")

    # Sample first center and compute proposal
    centers[0, :] = X[np.random.randint(X.shape[0])]
    distance_to_first = np.min(cdist(X, centers[0:1, :], metric='sqeuclidean'), axis=1)
    q = np.ascontiguousarray(.5 * distance_to_first/np.sum(distance_to_first) + .5 * 1/X.shape[0])

    # Taking view instead of python object
    cdef double[:, ::1] X_view = X
    cdef double[::1] q_view = q
    cdef double[:, ::1] centers_view = centers
    # cdef double[:, ::1] centers_view = np.zeros((k, X.shape[1]), np.float64, order="C")

    # Define Vose sampler
    cdef Sampler sampler
    sampler = Sampler(q_view)

    # Adding types for the loops
    cdef Py_ssize_t i, j
    cdef Py_ssize_t i_max = k - 1
    cdef Py_ssize_t j_max = chain_length - 1

    # Computing centroids
    for i in range(i_max):
        x_id = sampler.sample_1()
        x_position = X_view[x_id]
        x_proba = q_view[x_id]
        distance_to_x = min_sqeuclidean_distance(centers_view[:i + 1], x_position)

        # Starting the Markov Chain
        for j in range(j_max):
            y_id = sampler.sample_1()
            y_position = X_view[y_id]
            y_proba = q_view[y_id]
            distance_to_y = min_sqeuclidean_distance(centers_view[:i + 1], y_position)

            if distance_to_x * y_proba == 0:
                acceptance_ratio = 1
            else:
                acceptance_ratio = (distance_to_y * x_proba) / (distance_to_x * y_proba)
            if acceptance_ratio > (rand() / (RAND_MAX + 1.0)):
                x_position = y_position
                distance_to_x = distance_to_y
        centers_view[i+1] = x_position

    return np.asarray(centers_view)  #, weights